
/* MD5DEEP - sha1.c
 *
 * By Jesse Kornblum
 *
 * This is a work of the US Government. In accordance with 17 USC 105,
 * copyright protection is not available for any work of the US Government.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */



/* This implementation of SHA1-1 was written by 
   Steve Reid <steve@edmweb.com> and was included in this program on 
   the good faith that it is public domain */

/* Test Vectors (from FIPS PUB 180-1)
"abc"
  A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D
"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
  84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
A million repetitions of "a"
  34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
*/

#include "sha1.h"
#include <memory.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include "wvimage.h" //B64 funcs 


#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

/* blk0() and blk() perform the initial expand. */
/* I got the idea of expanding during the round function from SSLeay */
#if BYTE_ORDER == LITTLE_ENDIAN
#define blk0(i) (block->l[i] = (rol(block->l[i],24)&0xFF00FF00) \
    |(rol(block->l[i],8)&0x00FF00FF))
#else
#define blk0(i) block->l[i]
#endif
#define blk(i) (block->l[i&15] = rol(block->l[(i+13)&15]^block->l[(i+8)&15] \
    ^block->l[(i+2)&15]^block->l[i&15],1))

/* (R0+R1), R2, R3, R4 are the different operations used in SHA1 */
#define R0(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk0(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R1(v,w,x,y,z,i) z+=((w&(x^y))^y)+blk(i)+0x5A827999+rol(v,5);w=rol(w,30);
#define R2(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0x6ED9EBA1+rol(v,5);w=rol(w,30);
#define R3(v,w,x,y,z,i) z+=(((w|x)&y)|(w&x))+blk(i)+0x8F1BBCDC+rol(v,5);w=rol(w,30);
#define R4(v,w,x,y,z,i) z+=(w^x^y)+blk(i)+0xCA62C1D6+rol(v,5);w=rol(w,30);


/* Hash a single 512-bit block. This is the core of the algorithm. */

void SHA1Transform(unsigned long state[5], unsigned char buffer[SHA1_BLOCK])
{
  unsigned long a, b, c, d, e;
  typedef union {
    unsigned char c[SHA1_BLOCK];
    unsigned long l[16];
  } CHAR64LONG16;
  CHAR64LONG16* block;
#ifdef SHA1HANDSOFF
  static unsigned char workspace[SHA1_BLOCK];
  block = (CHAR64LONG16*)workspace;
  memcpy(block, buffer, SHA1_BLOCK);
#else
  block = (CHAR64LONG16*)buffer;
#endif
  /* Copy context->state[] to working vars */
  a = state[0];
  b = state[1];
  c = state[2];
  d = state[3];
  e = state[4];
  /* 4 rounds of 20 operations each. Loop unrolled. */
  R0(a,b,c,d,e, 0); R0(e,a,b,c,d, 1); R0(d,e,a,b,c, 2); R0(c,d,e,a,b, 3);
  R0(b,c,d,e,a, 4); R0(a,b,c,d,e, 5); R0(e,a,b,c,d, 6); R0(d,e,a,b,c, 7);
  R0(c,d,e,a,b, 8); R0(b,c,d,e,a, 9); R0(a,b,c,d,e,10); R0(e,a,b,c,d,11);
  R0(d,e,a,b,c,12); R0(c,d,e,a,b,13); R0(b,c,d,e,a,14); R0(a,b,c,d,e,15);
  R1(e,a,b,c,d,16); R1(d,e,a,b,c,17); R1(c,d,e,a,b,18); R1(b,c,d,e,a,19);
  R2(a,b,c,d,e,20); R2(e,a,b,c,d,21); R2(d,e,a,b,c,22); R2(c,d,e,a,b,23);
  R2(b,c,d,e,a,24); R2(a,b,c,d,e,25); R2(e,a,b,c,d,26); R2(d,e,a,b,c,27);
  R2(c,d,e,a,b,28); R2(b,c,d,e,a,29); R2(a,b,c,d,e,30); R2(e,a,b,c,d,31);
  R2(d,e,a,b,c,32); R2(c,d,e,a,b,33); R2(b,c,d,e,a,34); R2(a,b,c,d,e,35);
  R2(e,a,b,c,d,36); R2(d,e,a,b,c,37); R2(c,d,e,a,b,38); R2(b,c,d,e,a,39);
  R3(a,b,c,d,e,40); R3(e,a,b,c,d,41); R3(d,e,a,b,c,42); R3(c,d,e,a,b,43);
  R3(b,c,d,e,a,44); R3(a,b,c,d,e,45); R3(e,a,b,c,d,46); R3(d,e,a,b,c,47);
  R3(c,d,e,a,b,48); R3(b,c,d,e,a,49); R3(a,b,c,d,e,50); R3(e,a,b,c,d,51);
  R3(d,e,a,b,c,52); R3(c,d,e,a,b,53); R3(b,c,d,e,a,54); R3(a,b,c,d,e,55);
  R3(e,a,b,c,d,56); R3(d,e,a,b,c,57); R3(c,d,e,a,b,58); R3(b,c,d,e,a,59);
  R4(a,b,c,d,e,60); R4(e,a,b,c,d,61); R4(d,e,a,b,c,62); R4(c,d,e,a,b,63);
  R4(b,c,d,e,a,64); R4(a,b,c,d,e,65); R4(e,a,b,c,d,66); R4(d,e,a,b,c,67);
  R4(c,d,e,a,b,68); R4(b,c,d,e,a,69); R4(a,b,c,d,e,70); R4(e,a,b,c,d,71);
  R4(d,e,a,b,c,72); R4(c,d,e,a,b,73); R4(b,c,d,e,a,74); R4(a,b,c,d,e,75);
  R4(e,a,b,c,d,76); R4(d,e,a,b,c,77); R4(c,d,e,a,b,78); R4(b,c,d,e,a,79);
  /* Add the working vars back into context.state[] */
  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;
  state[4] += e;
  /* Wipe variables */
  a = b = c = d = e = 0;
}


/* SHA1Init - Initialize new context */

void SHA1Init(SHA1_CTX* context)
{
  /* SHA1 initialization constants */
  context->state[0] = 0x67452301;
  context->state[1] = 0xEFCDAB89;
  context->state[2] = 0x98BADCFE;
  context->state[3] = 0x10325476;
  context->state[4] = 0xC3D2E1F0;
  context->count[0] = context->count[1] = 0;
}


/* Run your data through this. */

void SHA1Update(SHA1_CTX* context, unsigned char* data, unsigned int len)
{
  unsigned int i, j;

  j = (context->count[0] >> 3) & 63;
  if ((context->count[0] += len << 3) < (len << 3)) context->count[1]++;
  context->count[1] += (len >> 29);
  if ((j + len) > 63) {
    memcpy(&context->buffer[j], data, (i = 64-j));
    SHA1Transform(context->state, context->buffer);
    for ( ; i + 63 < len; i += 64) {
      SHA1Transform(context->state, &data[i]);
    }
    j = 0;
  }
  else i = 0;
  memcpy(&context->buffer[j], &data[i], len - i);
}


/* Add padding and return the message digest. */

void SHA1Final(unsigned char digest[20], SHA1_CTX* context)
{
  unsigned long i, j;
  unsigned char finalcount[8];

  for (i = 0; i < 8; i++) {
    finalcount[i] = (unsigned char)((context->count[(i >= 4 ? 0 : 1)]
				     >> ((3-(i & 3)) * 8) ) & 255);  /* Endian independent */
  }
  SHA1Update(context, (unsigned char *)"\200", 1);
  while ((context->count[0] & 504) != 448) {
    SHA1Update(context, (unsigned char *)"\0", 1);
  }
  SHA1Update(context, finalcount, 8);  /* Should cause a SHA1Transform() */
  for (i = 0; i < 20; i++) {
    digest[i] = (unsigned char)
      ((context->state[i>>2] >> ((3-(i & 3)) * 8) ) & 255);
  }
  /* Wipe variables */
  i = j = 0;
  memset(context->buffer, 0, 64);
  memset(context->state, 0, 20);
  memset(context->count, 0, 8);
  memset(&finalcount, 0, 8);
#ifdef SHA1HANDSOFF  /* make SHA1Transform overwrite it's own static vars */
  SHA1Transform(context->state, context->buffer);
#endif
}

#define ipad 0x36
#define opad 0x5c

/* 
 * Encode a string using HMAC - see RFC-2104 for details.
 */
unsigned char *	 hmac_sha1(unsigned char *K,/* key */
						 int Klen,/* and it size */
						 unsigned char *T, /* text to encode */
						 int Tlen /* and its size */)		
{
    SHA1_CTX hash_ctx;
    unsigned char keyin[SHA1_BLOCK];
    unsigned char *step2;
    unsigned char step4[SHA1_LENGTH];
    unsigned char step5[SHA1_BLOCK + SHA1_LENGTH];
    unsigned char *hmac_out;
    unsigned char *c;
    int i;
    int j;

    SHA1Init(&hash_ctx);

    /* If the key is bigger than SHA1_BLOCK we need to hash it. */
    if (Klen > SHA1_BLOCK) 
	{
		SHA1Update(&hash_ctx, K, Klen);
		SHA1Final(keyin, &hash_ctx);
		Klen = SHA1_LENGTH;
    } 
	else 
	{
		memcpy(keyin, K, Klen);
    }

    step2 = (unsigned char *) malloc(Tlen + SHA1_BLOCK);

    c = keyin;
    for (i = 0; i < Klen; i++) 
	{
		step2[i] = (unsigned char)(*c ^ ipad);
		step5[i] = (unsigned char)(*c ^ opad);
		c++;
    }
    for (j = i; j < SHA1_BLOCK; j++) 
	{
		step2[j] = 0x36;
		step5[j] = opad;
    }

    memcpy(&step2[SHA1_BLOCK], T, Tlen);

    SHA1Init(&hash_ctx);
    SHA1Update(&hash_ctx, step2, SHA1_BLOCK + Tlen);
    SHA1Final(step4, &hash_ctx);

    memcpy(&step5[SHA1_BLOCK], step4, SHA1_LENGTH);

    hmac_out = (unsigned char *) malloc(SHA1_LENGTH+1);

    SHA1Init(&hash_ctx);
    SHA1Update(&hash_ctx, step5, SHA1_BLOCK + SHA1_LENGTH);
    SHA1Final(hmac_out, &hash_ctx);

    return hmac_out;
}


/* 
 * Functions to implement type SHA1 cancel locks.  Written by Fluffy, Ruler
 * of All Usenet, and placed into the public domain.  Meow.
 */



#define BUFFSIZE 41
/* 
 * Return a stripped cancel lock, that is, with the xx: prefix
 * removed, else NULL on failure.
 * type is set to the lock type, else zero on failure.
 */
char * lock_strip_alpha1( char *key,  char *type)
{
    char *c;
    char typetext[256], keyout[256];


    strcpy( (char*)typetext, "");
    strcpy((char*)keyout, "");

    while ((*key != ':') && (*key != '\0')) {
	strncat(typetext, key++, 1);
    }
    if (*key == '\0') {
	type = NULL;
    } else {
	key++;
	for (c = key; *c != '\0'; c++) {
	    if (*c == ':')
		*c = '\0';
	}

	strcpy(type, typetext);
	for (c = type; *c != '\0'; ++c) {
	    *c = (char)tolower( (unsigned char) *c);
	}
    }
    return (key);
}


/* 
 * Generate an SHA1 cancel key.
 */
char * sha_key1(char  *secret, int seclen, char *message, int msglen)
{
    char *cankey;
    unsigned char *hmacbuff;

    cankey = ( char *) malloc(BUFFSIZE);

    hmacbuff = hmac_sha1( (unsigned char *) secret, seclen, (unsigned char *)  message, msglen);
    strcpy( (char *) cankey, "sha1:");
	Encode64(hmacbuff, SHA1_LENGTH, (char *)cankey + 5, BUFFSIZE);
    return (cankey);
}

/* 
 * Generate an SHA1 cancel lock.
 */
char *	 sha_lock1( char *secret, int seclen,  char *message, int msglen)
{
     char *cankey, *canlock;
     unsigned char hmacbuff[SHA1_LENGTH];
     char junk[SHA1_LENGTH];
    SHA1_CTX hash_ctx;

    cankey = lock_strip_alpha1(sha_key1(secret, seclen, message, msglen), junk);

    canlock = ( char *) malloc(BUFFSIZE);
    SHA1Init(&hash_ctx);
    SHA1Update(&hash_ctx, (unsigned char*)cankey, strlen((char*)cankey));
    SHA1Final(hmacbuff, &hash_ctx);
    strcpy((char *) canlock, "sha1:");
	Encode64(hmacbuff, SHA1_LENGTH, (char *)canlock + 5, BUFFSIZE);

    return (canlock);
}

/* 
 * Verify an SHA cancel key against a cancel lock.
 * Returns 0 on success, nonzero on failure.
 */
int sha_verify1( char *key,  char *lock)
{
    unsigned char binkey[SHA1_LENGTH + 4];
    char templock[BUFFSIZE];
    unsigned char hmacbuff[SHA1_LENGTH];
    SHA1_CTX hash_ctx;
    int verified;

    /* Convert the key back into binary */
    //(void) b64_pton(key, binkey, (SHA1_LENGTH + 4));
	KWIKdecode64((char *)key, binkey, -1, (SHA1_LENGTH + 4));

    SHA1Init(&hash_ctx);
    SHA1Update(&hash_ctx, (unsigned char*) key, strlen((char*)key));
    SHA1Final(hmacbuff, &hash_ctx);
    Encode64(hmacbuff, SHA1_LENGTH, (char *)templock, BUFFSIZE);


    verified = strcmp((char*)templock, (char*)lock);

    return (verified);
}



